"""
Support for generic H5 files.

This attempts to use some common names to read in some very simple h5 files.

The FORESEE data in PubDAS motivated this module.
"""
from __future__ import annotations

from .core import H5Simple
