"""
Basic support for DAS-HDF5 a subset of CF (climate and forcasting).

This was created mainly for reading PoroTomo data from Brady Hotsprings.

More info on PoroTomo here:
https://github.com/openEDI/documentation/tree/main/PoroTomo
"""
from __future__ import annotations

from .core import DASHDF5
