"""
Support for prodML format.

This is used by Silixa's iDAS as an HDF5 format and perhaps other interrogators.

More info about ProdML can be found here:
https://www.energistics.org/prodml-developers-users
"""
from __future__ import annotations
from .core import ProdMLV2_0, ProdMLV2_1
